//! Subscription manager
use super::AllClients;
use crate::{client, Error};
use futures::future::join_all;
use nostr_rs_storage_base::Index;
use nostr_rs_types::{
    client::subscribe::{self, is_all_events},
    relayer,
    types::{event::SortedFilter, Filter, Index as EventIndex, SubscriptionId},
    Response,
};
use std::{
    collections::{BTreeMap, HashSet},
    sync::{
        atomic::{AtomicUsize, Ordering},
        Arc,
    },
};
use tokio::sync::{mpsc, RwLock};
use url::Url;

#[derive(Debug, Copy, Eq, PartialEq, Clone)]
/// Subscription status
pub enum Status {
    /// Subscription is awaiting to be subscribed
    Queued,
    /// Subscribed is active
    Subscribed,
    /// Technically unsubscribed, and fetching future events from the relayer
    /// from the All-Events meta subscription
    Stale,
}

#[allow(dead_code)]
struct Subscription {
    /// Request to subscribe
    subscription_request: subscribe::Subscribe,
    /// Active subscription (in the client side), when this is Drop all clients unsubscribes
    active_subscription: Option<Vec<client::ActiveSubscription>>,
    /// Subscription status
    status: Status,
    /// Reverse index
    ///
    /// This is a reverse index of the filters, it is only used to update the
    /// main shared index when this subscription is dropped.
    reverse_index: Vec<(Filter, Vec<EventIndex>)>,
}

/// Active subscription
///
/// This object is responsible for keeping track of a subscription for a
/// connection
///
/// This must be dropped to unsubscribe from the subscription manager
pub struct ActiveSubscription {
    id: (SubscriptionId, Option<Url>),
    manager: Arc<Manager>,
    active_subscriptions: Arc<AtomicUsize>,
    queued_subscriptions: Arc<AtomicUsize>,
    stale_subscriptions: Arc<AtomicUsize>,
}

impl Drop for ActiveSubscription {
    fn drop(&mut self) {
        let manager = self.manager.clone();
        let id_to_remove = self.id.clone();
        let active_subscriptions = self.active_subscriptions.clone();
        let queued_subscriptions = self.queued_subscriptions.clone();
        let stale_subscriptions = self.stale_subscriptions.clone();

        tokio::spawn(async move {
            let mut subscriptions = manager.subscriptions.write().await;
            let mut indexes = manager.index.write().await;

            if let Some(subscription) = subscriptions.remove(&id_to_remove) {
                match subscription.status {
                    Status::Subscribed => active_subscriptions.fetch_sub(1, Ordering::Relaxed),
                    Status::Queued => queued_subscriptions.fetch_sub(1, Ordering::Relaxed),
                    Status::Stale => stale_subscriptions.fetch_sub(1, Ordering::Relaxed),
                };

                for (_, single_indexes) in subscription.reverse_index.iter() {
                    for index in single_indexes.iter() {
                        indexes.remove(&(index.clone(), id_to_remove.0.clone()));
                    }
                }
            }
        });
    }
}

/// Subscription manager
///
/// Clients who are added to the pool are automatically subscribed to all
/// events, which is known as the All-Events subscription.
///
/// Subscriptions in the client pool are smarter than the raw subscriptions at
/// the client level. These subscriptions will be active until an "End of
/// stored" event is received; the client pool will unsubscribe at that point,
/// and the All-Events subscriptions, filtered internally, will fetch future
/// events. By doing so, only past events are queried, and there is a single
/// stream of future events to be a good citizen with other relayers.
#[derive(Default)]
pub(crate) struct Manager {
    subscriptions: RwLock<BTreeMap<(SubscriptionId, Option<Url>), Subscription>>,
    index: RwLock<BTreeMap<(EventIndex, SubscriptionId), SortedFilter>>,
    all_clients: AllClients,
    active_subscriptions: Arc<AtomicUsize>,
    queued_subscriptions: Arc<AtomicUsize>,
    stale_subscriptions: Arc<AtomicUsize>,
}

/// Maximum number of subscriptions
pub const MAX_SUBSCRIPTIONS: usize = 50;

impl Manager {
    /// Processes all messages from the client pools
    ///
    /// The client pool creates a subscription to the All-Events subscription,
    /// this callback checks if there are any listener to this event, otherwise
    /// it will not process the event.
    pub async fn process_message(
        self: &Arc<Self>,
        message: Response,
        url: Url,
        return_to: mpsc::Sender<(Response, Url)>,
    ) -> Result<(), Error> {
        match message {
            Response::EndOfStoredEvents(subscription_id) => {
                let subscription_id = (subscription_id.0, None);
                let mut subscription = self.subscriptions.write().await;
                subscription.get_mut(&subscription_id).map(|s| {
                    s.status = Status::Stale;
                    let _ = s.active_subscription.take();

                    self.active_subscriptions.fetch_sub(1, Ordering::Relaxed);
                    self.stale_subscriptions.fetch_add(1, Ordering::Relaxed);
                });

                return_to
                    .try_send((Response::EndOfStoredEvents(subscription_id.0.into()), url))
                    .map_err(|e| Error::InternalChannel(e.to_string()))?;

                return Ok(());
            }
            Response::Event(relayer::Event {
                subscription_id,
                event,
            }) => {
                if !is_all_events(&subscription_id) {
                    // This is not an All-Events subscription, it must be passed on as it is an stored event
                    return_to
                        .try_send((
                            Response::Event(relayer::Event {
                                subscription_id: subscription_id,
                                event,
                            }),
                            url.clone(),
                        ))
                        .map_err(|e| Error::InternalChannel(e.to_string()))?;
                    return Ok(());
                }

                let index = self.index.read().await;
                let mut matched = HashSet::new();
                let event_index = event.get_indexes(4);

                for idx in event_index {
                    let mut start = index.range(&(idx.clone(), SubscriptionId::empty())..);

                    while let Some(((current_idx, subscription_id), filter)) = start.next() {
                        if current_idx != &idx {
                            break;
                        }

                        if !matched.contains(subscription_id) && filter.check_event(&event) {
                            return_to
                                .try_send((
                                    Response::Event(relayer::Event {
                                        subscription_id: subscription_id.clone(),
                                        event: event.clone(),
                                    }),
                                    url.clone(),
                                ))
                                .map_err(|e| Error::InternalChannel(e.to_string()))?;
                            matched.insert(subscription_id.clone());
                        }
                    }
                }

                Ok(())
            }
            any_message => {
                return_to
                    .try_send((any_message, url))
                    .map_err(|e| Error::InternalChannel(e.to_string()))?;
                Ok(())
            }
        }
    }

    async fn update_active_subscriptions(self: &Arc<Self>) {
        if self.active_subscriptions.load(Ordering::Relaxed) >= MAX_SUBSCRIPTIONS
            || self.queued_subscriptions.load(Ordering::Relaxed) == 0
        {
            return;
        }

        let clients = self.all_clients.read().await;
        let mut subscriptions = self.subscriptions.write().await;

        for subscription in subscriptions.values_mut() {
            if subscription.status == Status::Queued {
                let wait_all = clients
                    .values()
                    .map(|(_, (_, sender))| {
                        sender.subscribe(subscription.subscription_request.clone())
                    })
                    .collect::<Vec<_>>();

                if let Ok(active_subscriptions) = join_all(wait_all)
                    .await
                    .into_iter()
                    .collect::<Result<Vec<_>, _>>()
                {
                    subscription.active_subscription = Some(active_subscriptions);
                    subscription.status = Status::Subscribed;

                    let queued_subscribed =
                        self.queued_subscriptions.fetch_sub(1, Ordering::Relaxed);
                    let active_subscriptions =
                        self.active_subscriptions.fetch_add(1, Ordering::Relaxed);
                    if queued_subscribed == 0 || active_subscriptions >= MAX_SUBSCRIPTIONS {
                        break;
                    }
                }
            }
        }
    }

    /// Creates a new subscription with a given filters
    pub async fn subcribe(
        self: &Arc<Self>,
        mut subscription_request: subscribe::Subscribe,
        specific_url: Option<Url>,
    ) -> ActiveSubscription {
        let mut subscriptions = self.subscriptions.write().await;
        let mut indexes = self.index.write().await;

        let id = (subscription_request.subscription_id.clone(), specific_url);
        let reverse_index: Vec<_> = subscription_request
            .filters
            .iter_mut()
            .map(|f| {
                let index = <&mut Filter as Into<Index>>::into(f).split();
                (f.clone(), index)
            })
            .collect();

        for (filter, single_indexes) in reverse_index.iter() {
            for index in single_indexes.iter() {
                indexes.insert((index.clone(), id.0.clone()), filter.clone().into());
            }
        }

        subscriptions.insert(
            id.clone(),
            Subscription {
                reverse_index,
                subscription_request,
                status: Status::Queued,
                active_subscription: None,
            },
        );

        self.queued_subscriptions.fetch_add(1, Ordering::Relaxed);

        let this = self.clone();
        tokio::spawn(async move {
            this.update_active_subscriptions().await;
        });

        ActiveSubscription {
            id,
            manager: self.clone(),
            active_subscriptions: self.active_subscriptions.clone(),
            queued_subscriptions: self.queued_subscriptions.clone(),
            stale_subscriptions: self.stale_subscriptions.clone(),
        }
    }

    /// Total number of subscribers
    pub fn total_subscribers(&self) -> usize {
        self.active_subscriptions.load(Ordering::Relaxed)
            + self.queued_subscriptions.load(Ordering::Relaxed)
            + self.stale_subscriptions.load(Ordering::Relaxed)
    }
}
